/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "egEvaluator.h"
#include "bit.h"
#include "square.h"

inline gboolean bishopsCannotMate(guint64 bishops){
	gboolean white = FALSE, black = FALSE;
	while(bishops){
		if(getFirstBitAndClear64(&bishops) % 2 == 0) white = TRUE;
		else black = TRUE;
		if(white && black) return FALSE;
	}
	return TRUE;
}

inline gint kingsDistance(const Position* position){
	return distance(getFirstBit64(position->piece[White_King]),getFirstBit64(position->piece[Black_King]));
}

inline gint KBBK(const Position * position){
	gint loneKingPlayer = position->turn;
	if(position->piece[Bishop+loneKingPlayer] != 0) loneKingPlayer ^= 1;

	if(bishopsCannotMate(position->piece[Bishop + (loneKingPlayer^1)])) return 0;

	gint score= 11.0*Pawn_Value-kingsDistance(position)*(Pawn_Value/8.0)-
				cornerDistance(getFirstBit64(position->piece[King + loneKingPlayer]))*
				(Pawn_Value/4.0);
				
	if(loneKingPlayer != (position->turn^1)) score = -score;
	return score;
}

inline gint KBNK(const Position * position){
 	gint loneKingPlayer = position->turn;
	if(position->piece[Bishop + loneKingPlayer] != 0) loneKingPlayer ^= 1;
	
	gint king = getFirstBit64(position->piece[King + loneKingPlayer]);
	gint bishop = getFirstBit64(position->piece[Bishop + (loneKingPlayer^1)]);
	gint f = file(bishop); 
	gint r = rank(bishop); 
	gint color = White;
	gint kb_cornerDistance;
	if((r%2 == 0 && f%2 == 0) || (r%2 != 0 && f%2 != 0)) color = Black;
	if(color == White) kb_cornerDistance = MIN(distance(king,h1),distance(king,a8));
	else kb_cornerDistance = MIN(distance(king,a1),distance(king,h8));
	gint score = 10.0*Pawn_Value - kb_cornerDistance*(Pawn_Value/4.0) - 		
				kingsDistance(position)*(Pawn_Value/8.0);
	if(loneKingPlayer != (position->turn ^ 1)) score = -score;
	return score;
}

inline gint KRK(const Position* position){
 	gint loneKingPlayer = position->turn;
	if(position->piece[Rook + loneKingPlayer] != 0) loneKingPlayer ^= 1;
	
	gint king =getFirstBit64(position->piece[King+loneKingPlayer]);
	gint f = file(king); 
	gint r = rank(king); 
	gint distance = MIN(MIN((7-f),f),MIN((7-r),r));
	gint score = 12.0*Pawn_Value - distance*(Pawn_Value/4.0)-kingsDistance(position)*(Pawn_Value/8.0);
	if(loneKingPlayer != (position->turn ^ 1)) score = -score;
	return score;
}

inline gint KQK(const Position* position){
 	gint loneKingPlayer = position->turn;
	if(position->piece[Queen + loneKingPlayer] != 0) loneKingPlayer ^= 1;
	
	gint king = getFirstBit64(position->piece[King+loneKingPlayer]);
	gint f = file(king); 
	gint r = rank(king); 
	gint distance = MIN(MIN((7-f),f),MIN((7-r),(r)));
	gint score = 13.0*Pawn_Value - distance*(Pawn_Value/4.0)-kingsDistance(position)*(Pawn_Value/8.0);
	if(loneKingPlayer != (position->turn ^ 1)) score = -score;
	return score;
}

gboolean egEvaluate(const Position* position, gint* eval){
	gint material = position->material[White] - position->material[Black];
	gint numberOfPieces = numberOfBits64(position->rotateL90);
	switch(numberOfPieces){
		case 2:{ *eval = 0; return TRUE;} /* KK */
		case 3:{
			switch(material){
				case Bishop_Value 	:
				case -Bishop_Value	:{ *eval = 0; return TRUE; }/* KBK */
				case Knight_Value	:
				case -Knight_Value	:{ *eval = 0; return TRUE; } /* KNK */
				case Rook_Value		:
				case -Rook_Value	:{ *eval = KRK(position); return TRUE; } /* KRK */
				case Queen_Value	:
				case -Queen_Value	:{ *eval = KQK(position); return TRUE; } /* KQK */
				case Pawn_Value		:
				case -Pawn_Value	:{ return FALSE; } /* KPK ??? */
			}
			break;
		}
		case 4:{
			switch(material){
				case(Bishop_Value + Bishop_Value)	:
				case(-(Bishop_Value+Bishop_Value))	:{ *eval = KBBK(position); return TRUE; } /* KBBK */
				case(Bishop_Value+Knight_Value)		:
				case(-(Bishop_Value+Knight_Value))	:{ *eval = KBNK(position); return TRUE; } /* KBNK */
				case(Knight_Value+Knight_Value)		:
				case(-(Knight_Value+Knight_Value))	:{ *eval = 0; return TRUE; } /* KNNK */ 
				case 0:{
					if(position->piece[White_Knight] != 0){
						   *eval = 0; /* KNKN */
						   return TRUE;
					}
					if(position->piece[White_Bishop] != 0){
							*eval = 0; /* KBKB */
							return TRUE;
					}
				}
				case(Knight_Value - Bishop_Value)	:
				case(Bishop_Value - Knight_Value)	:{ *eval = 0; return TRUE; } /* KBKN */
				case(Rook_Value - Knight_Value)		:
				case(-(Rook_Value - Knight_Value))	:{ *eval = 0; return TRUE; } /* KRKN */
				case(Rook_Value - Bishop_Value)		:
				case(-(Rook_Value - Bishop_Value))	:{ *eval = 0; return TRUE; } /* BRKB */
				case(Knight_Value - Pawn_Value)		:
				case(-(Knight_Value - Pawn_Value))	:{ *eval = 0; return TRUE; } /* KNKP */
				case(Bishop_Value - Pawn_Value)		:
				case(-(Bishop_Value - Pawn_Value))	:{ *eval = 0; return TRUE; } /* KBKP */
			}
			break;
		}
	}
	return FALSE;
}
