/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <glib/gprintf.h>

#include "evaluator.h"
#include "egEvaluator.h"
#include "startSearch.h"
#include "bit.h"
#include "piece.h"
#include "egbb.h"
#include "hash.h"
#include "search.h"

struct sevalInfo{
	gint kingSafety[2];
	gint majorMaterial[2];
	gint phase;
	gint evalOpening;
	gint evalEndgame;
	guint64 passedPawn[2];
	guint attackNumber[2];
	guint attackTotal[2];
};
typedef struct sevalInfo EvalInfo;

const gint PawnValueOpening = 70;
const gint KnightValueOpening = 325;
const gint BishopValueOpening = 325;
const gint RookValueOpening = 500;
const gint QueenValueOpening = 975;

const gint PawnValueEndgame = 90;
const gint KnightValueEndgame = 325;
const gint BishopValueEndgame = 325;
const gint RookValueEndgame = 500;
const gint QueenValueEndgame = 975;

const gint BishopPairOpening = 50;
const gint BishopPairEndgame = 50;

const gint TrappedBishop = 100;
const gint BlockedBishop = 50;
const gint BlockedRook = 50;
const gint BlockedCenterPawn = 10;

const gint KnightUnit = 4;
const gint BishopUnit = 6;
const gint RookUnit = 7;

const gint KnightMobOpening = 4;
const gint KnightMobEndgame = 4;
const gint BishopMobOpening = 5;
const gint BishopMobEndgame = 5;
const gint RookMobOpening = 2;
const gint RookMobEndgame = 4;
const gint QueenMobOpening = 1;
const gint QueenMobEndgame = 2;
const gint KingMobOpening = 0;
const gint KingMobEndgame = 0;
const gint RookSemiOpenFileOpening = 10;
const gint RookSemiOpenFileEndgame = 10;
const gint RookOpenFileOpening = 20;
const gint RookOpenFileEndgame = 20;
const gint RookSemiKingFileOpening = 10;
const gint RookKingFileOpening = 20;
const gint Rook7thOpening = 20;
const gint Rook7thEndgame = 40;
const gint Queen7thOpening = 10;
const gint Queen7thEndgame = 20;
const gint StormOpening = 10;
const gint KingAttackOpening = 20;
const gint DoubledOpening = 10;
const gint DoubledEndgame = 20;
const gint IsolatedOpening = 10;
const gint IsolatedOpeningOpen = 20;
const gint IsolatedEndgame = 20;
const gint BackwardOpening = 8;
const gint BackwardOpeningOpen = 16;
const gint BackwardEndgame = 10;
const gint CandidateOpeningMin = 5;
const gint CandidateOpeningMax = 55;
const gint CandidateEndgameMin = 10;
const gint CandidateEndgameMax = 110;
const gint PassedOpeningMin = 10;
const gint PassedOpeningMax = 70;
const gint PassedEndgameMin = 20;
const gint PassedEndgameMax = 140;
const gint UnstoppablePasser = 800;
const gint FreePasser = 60;
const gint AttackerDistance = 5;
const gint DefenderDistance = 20;
const gint ShelterOpening = 256;

const gint KingAttackUnit[14] = {0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 4, 4, 0 ,0};
const gint KingAttackWeight[16] = {0, 0, 128, 192, 224, 240, 248, 252, 254, 255, 256, 256 ,256, 256, 256, 256};
const gint Bonus[8] = {0,0,0,26,77,154,256,0};

const gint lazyEvalMargin = 200;

const gint KnightOutpostMatrix[2][64] = {
{	 
	0, 0, 0, 0, 0, 0, 0, 0,
   	0, 0, 0, 0, 0, 0, 0, 0,
   	0, 0, 0, 0, 0, 0, 0, 0,
   	0, 2, 5, 10, 10, 5, 2, 0,
   	0, 2, 5, 10, 10, 5, 2, 0,
   	0, 0, 4, 5, 5, 4, 0, 0,
   	0, 0, 0, 0, 0, 0, 0, 0,
   	0, 0, 0, 0, 0, 0, 0, 0
},
{
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 4, 5, 5, 4, 0, 0,
	0, 2, 5, 10, 10, 5, 2, 0,
	0, 2, 5, 10, 10, 5, 2, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0
}
};

void evalMaterial(const Position* position, EvalInfo* evalInfo);
void evalPatterns(const Position* position, EvalInfo* eInfo);
void evalPieces(const Position* position, EvalInfo* eInfo);
void evalKing(const Position* position, EvalInfo* eInfo);
void evalPawns(const Position* position, EvalInfo* eInfo);
void evalPasser(const Position* position, EvalInfo* eInfo);
inline gint shelterSquare(const Position* position, gint square, gint color);
inline guint shelterFile(const Position* position, guint file, guint rank, guint color);
inline guint stormFile(const Position * position, guint file, guint color);
inline gint quad(gint y_min, gint y_max, gint x);
inline gboolean unstoppablePasser(const Position * position, gint square, gint player);
inline gboolean kingPasser(const Position* position, gint square, gint player);
inline gint pawnAttDist(gint square, gint king, gint player);
inline gint pawnDefDist(gint square, gint king, gint player);

gint evaluate(Position* position, gint alpha, gint beta){
	gint eval;
	EvalInfo eInfo;

	/* Material */
	evalMaterial(position, &eInfo);
	/*g_printf("Material %d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/

	gint numberOfPieces = numberOfBits64(position->rotateL90);
	if(numberOfPieces <= 5){
		if(egbbIsLoaded() && egbbProbe(position, &eval)){
			if(eval == 0) return EVAL_DRAW;
			if(position->turn == Black){
				eInfo.evalOpening = -eval;
				eInfo.evalEndgame = -eval;
			}
			else{
				eInfo.evalOpening = eval;
				eInfo.evalEndgame = eval;
			}
		}
		else if(numberOfPieces <= 4 && egEvaluate(position, &eval)) return eval;
	}

	/* PVT */
	eInfo.evalOpening += position->pvt[Opening];
	eInfo.evalEndgame += position->pvt[Endgame];
	/*g_printf("PVT %d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/

	/* Pawns */
	evalPawns(position, &eInfo);
	/*g_printf("Pawns %d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/

	/* Draw ?? */

	/* Tempo */
	if(position->turn == White){
		eInfo.evalOpening += 20;
		eInfo.evalEndgame += 10;
	}
	else{
		eInfo.evalOpening -= 20;
		eInfo.evalEndgame -= 10;
	}
	/*g_printf("Tempo %d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/

	/* Patterns */
	evalPatterns(position, &eInfo);
	/*g_printf("Patterns %d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/
	
	/* Lazy cutoff */
    if(eInfo.phase > 6){
        eval = ((eInfo.evalOpening * (256 - eInfo.phase)) + (eInfo.evalEndgame * eInfo.phase)) / 256;
        if(position->turn == Black) eval = -eval;
        if(eval - lazyEvalMargin >= beta) return eval;
        if(eval + lazyEvalMargin <= alpha) return eval;
    }

	/* Pieces */
	evalPieces(position, &eInfo);
	/*g_printf("Pieces %d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/

	/* King */
	evalKing(position, &eInfo);
	/*g_printf("King %d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/

	/* Passer */
	evalPasser(position, &eInfo);
	/*g_printf("Passer:%d/%d\n",eInfo.evalOpening,eInfo.evalEndgame);*/

	/* Evaluate current phase */
	eval = ((eInfo.evalOpening * (256 - eInfo.phase)) + (eInfo.evalEndgame * eInfo.phase)) / 256;

	if(position->turn == Black) eval = -eval;
	/*g_printf("Eval:%d\n",eval);*/
	return eval;
}

void evalMaterial(const Position* position, EvalInfo* eInfo){
	gint evalOpening=0, evalEndgame=0, kingSafety[2], majorMaterial[2];

	/* init */
	guint wp = numberOfBits64(position->piece[White_Pawn]);
	guint wn = numberOfBits64(position->piece[White_Knight]);
   	guint wb = numberOfBits64(position->piece[White_Bishop]);
	guint wr = numberOfBits64(position->piece[White_Rook]);
	guint wq = numberOfBits64(position->piece[White_Queen]);
	guint bp = numberOfBits64(position->piece[Black_Pawn]);
	guint bn = numberOfBits64(position->piece[Black_Knight]);
   	guint bb = numberOfBits64(position->piece[Black_Bishop]);
	guint br = numberOfBits64(position->piece[Black_Rook]);
	guint bq = numberOfBits64(position->piece[Black_Queen]);

	/* Phase */
	gint phase = Total_Phase;

	phase -= wn * Knight_Phase;
	phase -= wb * Bishop_Phase;
	phase -= wr * Rook_Phase;
	phase -= wq * Queen_Phase;

	phase -= bn * Knight_Phase;
	phase -= bb * Bishop_Phase;
	phase -= br * Rook_Phase;
	phase -= bq * Queen_Phase;

   	if(phase < 0) phase = 0;
   	phase = (phase * 256 + (Total_Phase / 2)) / Total_Phase;

	/* King safety */
	if (bq >= 1 && bq+br+bb+bn >= 2) kingSafety[White] = TRUE;
	else kingSafety[White] = FALSE;

	if (wq >= 1 && wq+wr+wb+wn >= 2) kingSafety[Black] = TRUE;
	else kingSafety[Black] = FALSE;

	/* Material */
	majorMaterial[White] = wn*KnightValueOpening + wb*BishopValueOpening + 
								wr*RookValueOpening + wq*QueenValueOpening;

	evalOpening += majorMaterial[White] + wp*PawnValueOpening;
	majorMaterial[Black] = bn*KnightValueOpening + bb*BishopValueOpening + 
								br*RookValueOpening + bq*QueenValueOpening;
	evalOpening -= majorMaterial[Black] + bp*PawnValueOpening;

	evalEndgame += wn*KnightValueEndgame + wb*BishopValueEndgame + wr*RookValueEndgame + 
					wq*QueenValueEndgame + wp*PawnValueEndgame;
	evalEndgame -= bn*KnightValueEndgame + bb*BishopValueEndgame + br*RookValueEndgame + 
					bq*QueenValueEndgame + bp*PawnValueEndgame;

	/* Bishop pair (different colors ??)*/
	if (wb >= 2){
	       evalOpening += BishopPairOpening;
	       evalEndgame += BishopPairEndgame;
	}
	if (bb >= 2){
	       evalOpening -= BishopPairOpening;
	       evalEndgame -= BishopPairEndgame;
	}

	/* Store info */
	eInfo->kingSafety[White] = kingSafety[White];
	eInfo->kingSafety[Black] = kingSafety[Black];
	eInfo->majorMaterial[White] = majorMaterial[White];
	eInfo->majorMaterial[Black] = majorMaterial[Black];
	eInfo->phase = phase;
	eInfo->evalOpening = evalOpening;
	eInfo->evalEndgame = evalEndgame;
}

void evalPawns(const Position* position, EvalInfo* eInfo){
	guint64 pawns, t1,t2, key;
	gint eval[2];
	gint player, opponent, square, pawnFile, pawnRank, evalOpening[2]={0,0}, evalEndgame[2]={0,0},n;
	gboolean backward, doubled, passed, isolated, open, candidate;

	key = position->piece[White_Pawn]|position->piece[Black_Pawn];
	PawnHashData* pd;

   	if(pawnHashFind(key,&pd, eval, eInfo->passedPawn)){
		eInfo->evalOpening += eval[Opening];
		eInfo->evalEndgame += eval[Endgame];
		return;		
	}

	for(player = White; player <= Black; player++){
		opponent = player^1;
		pawns = position->piece[Pawn+player];
		eInfo->passedPawn[player] = 0;
		while(pawns){
			square = getFirstBitAndClear64(&pawns);
			pawnFile = file(square);
			pawnRank = rank(square);
			backward = doubled = isolated = open = passed = candidate= FALSE;

			if(bitsLE((BFILE[pawnFile] & position->piece[Pawn+player]),pawnRank)){
				doubled = TRUE;
			}
			t1 = getBitFile(position->piece[Pawn+player],pawnFile-1)|
					getBitFile(position->piece[Pawn+player],pawnFile+1);
			t2 = getBitFile(position->piece[White_Pawn] | position->piece[Black_Pawn],pawnFile);
			
			if(!t1){ /* OK */
				isolated = TRUE;
			}else{ /* OK */
				if(player == White && !bitsLEEQ(t1,pawnRank)){ /* OK */
					backward = TRUE;
					if((t1 & BRANK[pawnRank+1]) != 0){ /* OK */
						if(((t2 & BRANK[pawnRank+1]) | /* OK */
							((getBitFile(position->piece[Pawn+opponent],pawnFile-1) | 
							  getBitFile(position->piece[Pawn+opponent],pawnFile+1)) & 
							 (BRANK[pawnRank+1] | BRANK[pawnRank+2]))) == 0){
							backward = FALSE;
						}
					}
					else if(pawnRank == 1 && ((t1&BRANK[pawnRank+2]) != 0)){ /* OK */
						if(((t2 & (BRANK[pawnRank+1] | BRANK[pawnRank+2])) | /*OK */
							((getBitFile(position->piece[Pawn+opponent],pawnFile-1) | 
							  getBitFile(position->piece[Pawn+opponent],pawnFile+1)) & 
							 (BRANK[pawnRank+1] | BRANK[pawnRank+2] | BRANK[pawnRank+3]))) == 0)
							backward = FALSE;
					}
				}
				if(player == Black && !bitsGTEQ(t1,pawnRank)){ /* OK */
					backward = TRUE;
					if((t1 & BRANK[pawnRank-1]) != 0){ /* OK */
						if(((t2 & BRANK[pawnRank-1]) | /* OK */
							((getBitFile(position->piece[Pawn+opponent],pawnFile-1) | 
							  getBitFile(position->piece[Pawn+opponent],pawnFile+1)) & 
							 (BRANK[pawnRank-1] | BRANK[pawnRank-2]))) == 0 ){
							backward = FALSE;
						}
					}
					else if(pawnRank == 6 && ((t1&BRANK[pawnRank-2]) != 0)){ /* OK */
						if(((t2 & (BRANK[pawnRank-1] | BRANK[pawnRank-2])) | /* OK */
							((getBitFile(position->piece[Pawn+opponent],pawnFile-1) | 
							  getBitFile(position->piece[Pawn+opponent],pawnFile+1)) & 
							 (BRANK[pawnRank-1] | BRANK[pawnRank-2] | BRANK[pawnRank-3]))) == 0)
							backward = FALSE;
					}
				}
			}

			if(player == White && !bitsGT(t2, pawnRank)){ /* OK */
				open = TRUE;
				if(!bitsGT(getBitFile(position->piece[Pawn+opponent], pawnFile-1),pawnRank) && /* OK */
					!bitsGT(getBitFile(position->piece[Pawn+opponent],pawnFile+1),pawnRank)){
					passed = TRUE;
					eInfo->passedPawn[player] |= 1ull<<square;
				}
				else{
					n = 0;
					n += getNumberBitsLEEQ(getBitFile(position->piece[Pawn+player], pawnFile-1),pawnRank);
					n += getNumberBitsLEEQ(getBitFile(position->piece[Pawn+player], pawnFile+1),pawnRank);
					n -= getNumberBitsGT(getBitFile(position->piece[Pawn+opponent], pawnFile-1),pawnRank);
					n -= getNumberBitsGT(getBitFile(position->piece[Pawn+opponent], pawnFile+1),pawnRank);
					if(n>=0){ /* OK */
						n = 0;
						n += getNumberBitsEQ(getBitFile(position->piece[Pawn+player], pawnFile-1),pawnRank-1);
						n += getNumberBitsEQ(getBitFile(position->piece[Pawn+player], pawnFile+1),pawnRank-1);
						n -= getNumberBitsEQ(getBitFile(position->piece[Pawn+opponent], pawnFile-1),pawnRank+1);
						n -= getNumberBitsEQ(getBitFile(position->piece[Pawn+opponent], pawnFile+1),pawnRank+1);
						if(n >= 0)	candidate = TRUE;
					}
				}
			}
			if(player == Black && !bitsLE(t2, pawnRank)){ /* OK */
				open = TRUE;
				if(!bitsLE(getBitFile(position->piece[Pawn+opponent], pawnFile-1),pawnRank) && /* OK */
					!bitsLE(getBitFile(position->piece[Pawn+opponent],pawnFile+1),pawnRank)){
					passed = TRUE;
					eInfo->passedPawn[player] |= 1ull<<square;
				}else{
					n = 0;
					n += getNumberBitsGTEQ(getBitFile(position->piece[Pawn+player], pawnFile-1),pawnRank);
					n += getNumberBitsGTEQ(getBitFile(position->piece[Pawn+player], pawnFile+1),pawnRank);
					n -= getNumberBitsLE(getBitFile(position->piece[Pawn+opponent], pawnFile-1),pawnRank);
					n -= getNumberBitsLE(getBitFile(position->piece[Pawn+opponent], pawnFile+1),pawnRank);
					if(n >= 0){ /* OK */
						n = 0;
						n += getNumberBitsEQ(getBitFile(position->piece[Pawn+player], pawnFile-1),pawnRank+1);
						n += getNumberBitsEQ(getBitFile(position->piece[Pawn+player], pawnFile+1),pawnRank+1);
						n -= getNumberBitsEQ(getBitFile(position->piece[Pawn+opponent], pawnFile-1),pawnRank-1);
						n -= getNumberBitsEQ(getBitFile(position->piece[Pawn+opponent], pawnFile+1),pawnRank-1);
						if(n >= 0)	candidate = TRUE;
					}
				}
			}

			if(doubled){
            	evalOpening[player] -= DoubledOpening;
				evalEndgame[player] -= DoubledEndgame;
       		}

        	if (isolated){
       	    	if (open){
					evalOpening[player] -= IsolatedOpeningOpen;
       				evalEndgame[player] -= IsolatedEndgame;
    	    	}else{
					evalOpening[player] -= IsolatedOpening;
					evalEndgame[player] -= IsolatedEndgame;
        		}
        	}

			if (backward){
	            if (open){
					evalOpening[player] -= BackwardOpeningOpen;
					evalEndgame[player] -= BackwardEndgame;
				}else{
					evalOpening[player] -= BackwardOpening;
					evalEndgame[player] -= BackwardEndgame;
	            }
			}

			if (candidate) {
				if(player == White){
					evalOpening[player] += quad(CandidateOpeningMin,CandidateOpeningMax,pawnRank);
    	        			evalEndgame[player] += quad(CandidateEndgameMin,CandidateEndgameMax,pawnRank);
				}else{
					evalOpening[player] += quad(CandidateOpeningMin,CandidateOpeningMax,7-pawnRank);
    	        			evalEndgame[player] += quad(CandidateEndgameMin,CandidateEndgameMax,7-pawnRank);
				}
			}
		}
	}

	eval[Opening] = evalOpening[White] - evalOpening[Black];
	eval[Endgame] = evalEndgame[White] - evalEndgame[Black];
   	pawnHashWrite(key, pd, eval, eInfo->passedPawn);
	eInfo->evalOpening += eval[Opening];
	eInfo->evalEndgame += eval[Endgame];
}

inline gint quad(gint y_min, gint y_max, gint x) {
   gint y;
   y = y_min + ((y_max - y_min) * Bonus[x] + 128) / 256;
   return y;
}

void evalPatterns(const Position* position, EvalInfo* eInfo){
    /* Blocked center pawns */
    if(position->type[d2] == White_Pawn && position->type[d3] != Empty)
        eInfo->evalOpening -= BlockedCenterPawn;
    if(position->type[e2] == White_Pawn && position->type[e3] != Empty)
        eInfo->evalOpening -= BlockedCenterPawn;
    if(position->type[d7] == Black_Pawn && position->type[d6] != Empty)
        eInfo->evalOpening += BlockedCenterPawn;
    if(position->type[e7] == Black_Pawn && position->type[e6] != Empty)
        eInfo->evalOpening += BlockedCenterPawn;

	/* Trapped bishop (7th rank) */
	if((position->type[a7] == White_Bishop && position->type[b6] == Black_Pawn) ||
	   (position->type[b8] == White_Bishop && position->type[c7] == Black_Pawn)){
		eInfo->evalOpening -= TrappedBishop;
		eInfo->evalEndgame -= TrappedBishop;
	}
	if((position->type[h7] == White_Bishop && position->type[g6] == Black_Pawn) ||
	   (position->type[g8] == White_Bishop && position->type[f7] == Black_Pawn)){
		eInfo->evalOpening -= TrappedBishop;
		eInfo->evalEndgame -= TrappedBishop;
	}
	if((position->type[a2] == Black_Bishop && position->type[b3] == White_Pawn) ||
	   (position->type[b1] == Black_Bishop && position->type[c2] == White_Pawn)){
		eInfo->evalOpening += TrappedBishop;
		eInfo->evalEndgame += TrappedBishop;
	}
	if((position->type[h2] == Black_Bishop && position->type[g3] == White_Pawn) ||
	   (position->type[g1] == Black_Bishop && position->type[f2] == White_Pawn)){
		eInfo->evalOpening += TrappedBishop;
		eInfo->evalEndgame += TrappedBishop;
	}

	/* Trapped bishop (6th rank) */
	if (position->type[a6] == White_Bishop && position->type[b5] == Black_Pawn) {
      eInfo->evalOpening -= TrappedBishop / 2;
      eInfo->evalEndgame -= TrappedBishop / 2;
   }
   if (position->type[h6] == White_Bishop && position->type[g5] == Black_Pawn) {
      eInfo->evalOpening -= TrappedBishop / 2;
      eInfo->evalEndgame -= TrappedBishop / 2;
   }
   if (position->type[a3] == Black_Bishop && position->type[b4] == White_Pawn) {
      eInfo->evalOpening += TrappedBishop / 2;
      eInfo->evalEndgame += TrappedBishop / 2;
   }
   if (position->type[h3] == Black_Bishop && position->type[g4] == White_Pawn) {
      eInfo->evalOpening += TrappedBishop / 2;
      eInfo->evalEndgame += TrappedBishop / 2;
   }

	/* Blocked bishop */
	if(position->type[d2] == White_Pawn && position->type[d3] != Empty && position->type[c1] == White_Bishop)
		eInfo->evalOpening -= BlockedBishop;
	if(position->type[e2] == White_Pawn && position->type[e3] != Empty && position->type[f1] == White_Bishop)
		eInfo->evalOpening -= BlockedBishop;
	if(position->type[d7] == Black_Pawn && position->type[d6] != Empty && position->type[c8] == Black_Bishop)
		eInfo->evalOpening += BlockedBishop;
	if(position->type[e7] == Black_Pawn && position->type[e6] != Empty && position->type[f8] == Black_Bishop)
		eInfo->evalOpening += BlockedBishop;

	/* Blocked rook */
	if((position->type[c1]==White_King || position->type[b1]==White_King) &&
	   (position->type[a1]==White_Rook || position->type[a2]==White_Rook || position->type[b1]==White_Rook))
		eInfo->evalOpening -= BlockedRook;
	if((position->type[f1]==White_King || position->type[g1] == White_King) &&
	   (position->type[h1]==White_Rook || position->type[h2]==White_Rook || position->type[g1]==White_Rook))
		eInfo->evalOpening -= BlockedRook;
	if((position->type[c8]==Black_King || position->type[b8]==Black_King) &&
	   (position->type[a8]==Black_Rook || position->type[a7]==Black_Rook || position->type[b8]==Black_Rook))
		eInfo->evalOpening += BlockedRook;
	if((position->type[f8]==Black_King || position->type[g8]==Black_King) &&
	   (position->type[h8]==Black_Rook || position->type[h7]==Black_Rook || position->type[g8]==Black_Rook))
		eInfo->evalOpening += BlockedRook;
}

void evalPieces(const Position* position, EvalInfo* eInfo){
	guint64 pieces, freeSquares, attack, mobAttack, kingSafeSquares1, kingSafeSquares2;
	gint player, opponent=0, from, pieceFile, pieceRank, distance;
	gint kingFile, kingRank;
	gint evalOpening[2], evalEndgame[2], mobility;

	freeSquares = ~(position->piece[White]|position->piece[Black]);
	for(player=White; player<=Black; player++){
		opponent = !player;
		evalOpening[player] = evalEndgame[player] = 0;
		eInfo->attackNumber[opponent] = 0;
		eInfo->attackTotal[opponent] = 0;

		mobAttack = freeSquares | position->piece[opponent];
		from = getFirstBit64(position->piece[King+opponent]);
		kingFile = file(from);
		kingRank = rank(from);
		kingSafeSquares1 = kingSafeDistance1[from];
		kingSafeSquares2 = kingSafeDistance2[from];
	
		/* Knight */
		pieces = position->piece[Knight+player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			attack = attackKnight[from];
			/* King attack */
			if(attack & kingSafeSquares1){
				eInfo->attackNumber[opponent] ++;
				eInfo->attackTotal[opponent] += KingAttackUnit[Knight];
			}	
			/* Mobility */
			mobility = -KnightUnit + numberOfBits64(attack & mobAttack);
			evalOpening[player] += mobility*KnightMobOpening;
			evalEndgame[player] += mobility*KnightMobEndgame;
			/* Outpost */
			if(player == White){
				if(position->type[from-7] == White_Pawn){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
				if(position->type[from-9] == White_Pawn){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
			}else{
				if(position->type[from+7] == Black_Pawn){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
				if(position->type[from+9] == Black_Pawn){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
			}
		}
		/* Bishop */
		pieces = position->piece[Bishop+player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			attack = positionAttackBishop(position,from);
			/* King attack */
			if((maskBishop[from] & kingSafeSquares1) && (attack & kingSafeSquares2)){
				eInfo->attackNumber[opponent] ++;
				eInfo->attackTotal[opponent] += KingAttackUnit[Bishop];
			}
			/* Mobility */
			mobility = -BishopUnit + numberOfBits64(attack & mobAttack);
			evalOpening[player] += mobility * BishopMobOpening;
			evalEndgame[player] += mobility * BishopMobEndgame;
		}
		/* Rook */
		pieces = position->piece[Rook+player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			attack = positionAttackRook(position,from);
			/* King attack */
			if((maskRook[from] & kingSafeSquares1) && (attack & kingSafeSquares2)){
				eInfo->attackNumber[opponent] ++;
				eInfo->attackTotal[opponent] += KingAttackUnit[Rook];
			}
			/* Mobility */
			mobility = -RookUnit + numberOfBits64(attack & mobAttack);
			evalOpening[player] += mobility * RookMobOpening;
			evalEndgame[player] += mobility * RookMobEndgame;
			/* Open file */
			pieceFile = file(from);
			evalOpening[player] -= RookOpenFileOpening / 2;
			evalEndgame[player] -= RookOpenFileEndgame / 2;
			attack = BFILE[pieceFile];
			if((attack & position->piece[Pawn+player]) == 0){ /* No friendly pawns*/
				evalOpening[player] += RookSemiOpenFileOpening;
				evalEndgame[player] += RookSemiOpenFileEndgame;
				if((attack & position->piece[Pawn+opponent]) == 0){ /* No enemy pawns */
					evalOpening[player] +=  RookOpenFileOpening - RookSemiOpenFileOpening;
					evalEndgame[player] +=  RookOpenFileEndgame - RookSemiOpenFileEndgame;
				}
				if(eInfo->kingSafety[opponent]){
					distance = ABS(pieceFile-kingFile);
					if(distance <= 1){
						evalOpening[player] += RookSemiKingFileOpening;
						if(distance == 0){
							evalOpening[player] += 
								RookKingFileOpening - RookSemiKingFileOpening;
						}
					}
				}
			}
			/* 7th rank */
			pieceRank = rank(from);
			if(pieceRank == Get_Rank_7[player]){
				/* Opponent pawns on 7th rank or the king on 8th rank */
				if(position->piece[Pawn+opponent] & GET_RANK_7[player] || 
					kingRank == Get_Rank_8[player]){
						evalOpening[player] += Rook7thOpening;
						evalEndgame[player] += Rook7thEndgame;
				}
			}
		}
		/* Queen */
		pieces = position->piece[Queen+player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			/* King attack */
			if((maskQueen[from] & kingSafeSquares1) && 
				(positionAttackQueen(position,from) & kingSafeSquares2)){
				eInfo->attackNumber[opponent] ++;
				eInfo->attackTotal[opponent] += KingAttackUnit[Queen];
			}
			/* Distance from king */
			pieceFile = file(from);
			pieceRank = rank(from);
			mobility = 10 -(ABS(kingRank-pieceRank)+ABS(kingFile-pieceFile));
			evalOpening[player] += mobility;
			evalEndgame[player] += mobility;
			/* 7th rank */
			if(pieceRank == Get_Rank_7[player]){
				/* Opponent pawns on 7th rank or the king on rank 8 */
				if(position->piece[Pawn+opponent] & GET_RANK_7[player] || 
					kingRank == Get_Rank_8[player]){
						evalOpening[player] += Queen7thOpening;
						evalEndgame[player] += Queen7thEndgame;
				}
			}
		}
	}
	eInfo->evalOpening += (evalOpening[White] - evalOpening[Black]);
	eInfo->evalEndgame += (evalEndgame[White] - evalEndgame[Black]);
}

void evalKing(const Position* position, EvalInfo* eInfo){
	guint kingIndex, kingFile, penalty1,penalty2,tmp, player;
	gint evalOpening[2]={0,0},evalEndgame[2]={0,0}, penalty;
	
    /* White king safety patterns */
    if(eInfo->kingSafety[White]){
        if(position->type[g1] == White_King || position->type[h1] == White_King){
            if(position->type[g2] == White_Pawn &&
                (position->type[h2] == White_Pawn || position->type[h3] == White_Pawn)){
                    eInfo->kingSafety[White] = FALSE;
            }
            else if(position->type[f2] == White_Pawn &&
                position->type[g3] == White_Pawn && position->type[h2] == White_Pawn){
                    eInfo->kingSafety[White] = FALSE;
            }
        }
        else if(position->type[b1] == White_King || position->type[a1] == White_King){
            if(position->type[b2] == White_Pawn &&
                (position->type[a2] == White_Pawn || position->type[a3] == White_Pawn)){
                    eInfo->kingSafety[White] = FALSE;
              }
              else if(position->type[c2] == White_Pawn &&
                position->type[b3] == White_Pawn && position->type[a2] == White_Pawn){
                    eInfo->kingSafety[White] = FALSE;
            }
        }
    }

    /* Black king safety patterns */
    if(eInfo->kingSafety[Black]){
        if (position->type[g8] == Black_King || position->type[h8] == Black_King){
            if(position->type[g7] == Black_Pawn &&
                (position->type[h7] == Black_Pawn || position->type[h6] == Black_Pawn)){
                    eInfo->kingSafety[Black] = FALSE;
            }
            else if(position->type[f7] == Black_Pawn &&
                position->type[g6] == Black_Pawn && position->type[h7] == Black_Pawn){
                    eInfo->kingSafety[Black] = FALSE;
            }
        }
        else if(position->type[b8] == Black_King || position->type[a8] == Black_King){
            if(position->type[b7] == Black_Pawn &&
                (position->type[a7] == Black_Pawn || position->type[a6] == Black_Pawn)){
                    eInfo->kingSafety[Black] = FALSE;
            }
            else if(position->type[c7] == Black_Pawn &&
                position->type[b6] == Black_Pawn && position->type[a7] == Black_Pawn){
                    eInfo->kingSafety[Black] = FALSE;
            }
        }
    }

	/* White king */
	penalty = 0;
	kingIndex = getFirstBit64(position->piece[White_King]);
	kingFile = file(kingIndex);

	if(eInfo->kingSafety[White] && position->piece[Black_Queen]  != 0){
		penalty1 = shelterSquare(position,kingIndex,White);

		/* Castling */
		penalty2 = penalty1;
		if((position->castle & WHITE_KING_CASTLE) != 0){
			tmp = shelterSquare(position,g1,White);
			if(tmp < penalty2) penalty2 = tmp;
		}
		if((position->castle & WHITE_QUEEN_CASTLE) != 0){
			tmp = shelterSquare(position,b1,White);
			if(tmp < penalty2) penalty2 = tmp;
		}
		penalty = (penalty1+penalty2)/2;

	}
	penalty += stormFile(position,kingFile,White);
	if (kingFile != File_A) penalty += stormFile(position,kingFile-1,White);
	if (kingFile != File_H) penalty += stormFile(position,kingFile+1,White);

	evalOpening[White] -= (penalty * ShelterOpening) / 256;
	/*g_printf("King white %d\n",evalOpening[White]);*/

	/* Black king */
	penalty=0;
	kingIndex = getFirstBit64(position->piece[Black_King]);
	kingFile = file(kingIndex);
	if(eInfo->kingSafety[Black] && position->piece[White_Queen] != 0){
		penalty1 = shelterSquare(position,kingIndex,Black);

		/* Castling */
		penalty2 = penalty1;
		if((position->castle & BLACK_KING_CASTLE) != 0){
			tmp = shelterSquare(position,g8,Black);
			if(tmp <  penalty2) penalty2 = tmp;
		}
		if((position->castle & BLACK_QUEEN_CASTLE) != 0){
			tmp = shelterSquare(position,b8,Black);
			if(tmp <  penalty2) penalty2 = tmp;
		}
		penalty = (penalty1+penalty2)/2;
	}
	penalty += stormFile(position,kingFile,Black);
	if (kingFile != File_A) penalty += stormFile(position,kingFile-1,Black);
	if (kingFile != File_H) penalty += stormFile(position,kingFile+1,Black);

	evalOpening[Black] -= (penalty * ShelterOpening) / 256;
	/*g_printf("King black %d\n",evalOpening[Black]);*/
	/* King attack */
	for(player=White; player <=Black; player++){
/*		if(eInfo->kingSafety[player]){*/
			evalOpening[player] -= (eInfo->attackTotal[player] * KingAttackOpening * 
										KingAttackWeight[eInfo->attackNumber[player]]) / 256;
/*		}*/
	}
	eInfo->evalOpening += evalOpening[White] - evalOpening[Black];
	eInfo->evalEndgame += evalEndgame[White] - evalEndgame[Black];
}

inline gint shelterSquare(const Position* position, gint square, gint color){
	guint penalty=0, file, rank;

	file = file(square);
	rank = rank(square);
	penalty += shelterFile(position,file,rank,color) * 2;
	if (file != File_A) penalty += shelterFile(position,file-1,rank,color);
	if (file != File_H) penalty += shelterFile(position,file+1,rank,color);

	if (penalty == 0) penalty = 11; /* Weak back rank */

	penalty += stormFile(position,file,color);
	if (file != File_A) penalty += stormFile(position,file-1,color);
	if (file != File_H) penalty += stormFile(position,file+1,color);

	return penalty;
}

inline guint shelterFile(const Position* position, guint file, guint rank, guint color) {
	guint distance, penalty;
	guint64 pieces;
	
	pieces = position->piece[Pawn+color] & BFILE[file];
	
	if(color == White){
		distance = rank*8;
		pieces = (pieces >> distance)<<distance;
		if(pieces) distance = 7 - rank(getFirstBit64(pieces));
		else distance = 0;
	}
	else{
		distance = (7-rank)*8;
		pieces = (pieces << distance)>>distance;
		if(pieces) distance = rank(getLastBit64(pieces));
		else distance = 0;
	}

	penalty = 36 - distance * distance;
	return penalty;
}

inline guint stormFile(const Position * position, guint file, guint color) {
	guint distance, penalty = 0;
	guint64 pieces;

	pieces = position->piece[Pawn+(color^1)] & BFILE[file];
	if(color == White){
		if(pieces) distance = 7 - rank(getFirstBit64(pieces));
		else distance = 0;
	}
	else{
		if(pieces) distance = rank(getLastBit64(pieces));
		else distance = 0;
	}

	switch (distance) {
		case Rank_4:{ penalty = StormOpening * 1; break;}
		case Rank_5:{ penalty = StormOpening * 2; break;}
		case Rank_6:{ penalty = StormOpening * 3; break;}
	}
	return penalty;
}

void evalPasser(const Position* position, EvalInfo* eInfo){
		gint evalOpening[2] = {0,0}, evalEndgame[2] = {0,0}, min, max, delta;
		guint square, king, oppKing, player, opponent, pRank/*, pFile*/;
		guint64 passedPawn/*, pawns*/;

		king = getFirstBit64(position->piece[White_King]);
		oppKing = getFirstBit64(position->piece[Black_King]);
		
		for(player = White; player <= Black; player++){
			passedPawn = eInfo->passedPawn[player];
			opponent = !player;
			if(player==Black){
				delta = king;
				king = oppKing;
				oppKing = delta;
			}
			square = 64;
			while(passedPawn){
				square = getFirstBitAndClear64(&passedPawn);
				if(player == White) pRank = rank(square);
				else pRank = 7 - rank(square);
				evalOpening[player] += quad(PassedOpeningMin,PassedOpeningMax, pRank);
				
				min = PassedEndgameMin;
				max = PassedEndgameMax;
				delta = max - min;

				if(eInfo->majorMaterial[opponent] == 0 && 
					(unstoppablePasser(position,square,player) || kingPasser(position,square,player))){
						delta += UnstoppablePasser;
				}
				/*else {
						pFile = file(square);
						pawns =  position->piece[Pawn+player];
						if(pFile != File_H && BFILE[pFile+1] & pawns) 
							delta += FreePasser;
						if(pFile != File_A && BFILE[pFile-1] & pawns) 
							delta += FreePasser;
				}*/
				
				delta -= pawnAttDist(square,king,player) * AttackerDistance;
				delta += pawnDefDist(square,oppKing,player) * DefenderDistance;

				evalEndgame[player] += min;
				if (delta > 0) evalEndgame[player] += quad(0,delta,pRank);
			}
		}
		
		eInfo->evalOpening += (evalOpening[White] - evalOpening[Black]);
		eInfo->evalEndgame += (evalEndgame[White] - evalEndgame[Black]);
}

inline gboolean unstoppablePasser(const Position * position, gint square, gint player){
		gint pRank, oppKing, dist;

		pRank = rank(square);

		if(player == White){ /* two squares up */
			oppKing = getFirstBit64(position->piece[Black_King]);
			if(bitsGT(BFILE[file(square)] & position->piece[White],pRank)) /* Clear promotion path? */
				return FALSE;
			if(pRank == 1){
				pRank ++;
				square += 8;
			}
			dist = distance(oppKing,square+(8*(7-pRank)));
			if(position->turn == Black) dist --;
			if(dist > 7 - pRank) return TRUE;

		}
		else{
			oppKing = getFirstBit64(position->piece[White_King]);
			if(bitsLE(BFILE[file(square)] & position->piece[Black],pRank)) /* Clear promotion path? */
				return FALSE;
			if(pRank == 6){ /* two squares down */
				pRank --;
				square -= 8;
			}
			dist = distance(oppKing,square-(8*pRank));
			if(position->turn == White) dist --;
			if(dist > pRank) return TRUE;
		}
		return FALSE;
}

inline gboolean kingPasser(const Position* position, gint square, gint player) {
	guint king, pFile, pRank, promotion;

	king =  getFirstBit64(position->piece[King+player]);
	pFile = file(square);
	pRank = rank(square);

	if(player == White) promotion = king + (8*(7-rank(king)));
	else promotion = king - (8*rank(king));
	
	if (distance(king,promotion) <= 1 && distance(king,square) == 1 && 
		(file(king) != pFile  || (pFile != 0 && pFile != 7))){
			return TRUE;
	}
	return FALSE;
}

inline gint pawnAttDist(gint square, gint king, gint player) {
	gint pawn;
	if(player == White) pawn = square + 8;
	else pawn = square - 8;
	return distance(pawn,king);
}

inline gint pawnDefDist(gint square, gint king, gint player) {
	gint pawn;
	if(player == White) pawn = square + 8;
	else pawn = square - 8;
	return distance(pawn,king);
}

