/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "init.h"
#include "piece.h"
#include "square.h"
#include "position.h"
#include "bit.h"
#include "search.h"
#include "threadSearch.h"
#include "uci.h"

const gint Pawn_File[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint Pawn_Rank[8] = {+0, +0, +0, +0, +2, +4, +8, +0};
const gint Knight_Line[8] = {-4, -2, +0, +1, +1, +0, -2, -4};
const gint Knight_Rank[8] = {-2, -1, +0, +1, +2, +3, +2, +1};
const gint Bishop_Line[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint Rook_File[8] = {-2, -1, +0, +1, +1, +0, -1, -2};
const gint Queen_Line[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint King_File[8] = {+3, +4, +2, +0, +0, +2, +4, +3};
const gint King_Rank[8] = {+1, +0, -2, -3, -4, -5, -6, -7};
const gint King_Line[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint Pawn_File_Opening = 5;
const gint Pawn_Rank_Endgame = 2;
const gint Knight_Center_Opening = 5;
const gint Knight_Rank_Opening = 5;
const gint Knight_Center_Endgame = 5;
const gint Knight_Back_Rank_Opening = 0;
const gint Knight_Trapped = 100;
const gint Bishop_Center_Opening = 2;
const gint Bishop_Center_Endgame = 3;
const gint Bishop_Back_Rank_Opening = 10;
const gint Bishop_Diagonal_Opening = 4;
const gint Rook_File_Opening = 3;
const gint Queen_Center_Opening = 0;
const gint Queen_Center_Endgame = 4;
const gint Queen_Back_Rank_Opening = 5;
const gint King_File_Opening = 10;
const gint King_Rank_Opening = 10;
const gint King_Center_Endgame = 12;

guint64 squareKey[14][64];
guint64 enPassantKey[8];
guint64 castleKey[4];
guint64 turnKey;
gint pvt[14][64][2];
guint64 attackPawn[2][64];
guint64 attackKnight[64];
guint64 attackBishopL45[64][64];
guint64 attackBishopR45[64][64];
guint64 attackRookL00[64][64];
guint64 attackRookL90[64][64];
guint64 attackKing[64];
guint64 kingSafeDistance1[64];
guint64 kingSafeDistance2[64];
guint64 maskKnight[64];
guint64 maskBishop[64];
guint64 maskRook[64];
guint64 maskQueen[64];
guint8 bitsIn16bits[0x1u << 16];

void initZobrist(GRand* rand);
void initPvt(void);
void initAttack(void);
void initBitsIn16bits(void);
void initUCIOptions(void);

void init(GRand* mrand){
	initBitsIn16bits();
	initZobrist(mrand);
	initAttack();
	initPvt();
}

void initZobrist(GRand* mrand){
    gint i,j;
    for(i=0; i<14; i++)
        for(j=0; j<64; j++)
            squareKey[i][j] = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);

	for(i=0; i<8; i++)
            enPassantKey[i] = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);
	
	for(i=0; i<4; i++)
            castleKey[i] = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);

	turnKey = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);
}

void initPvt(void){
	gint i,j,k;
	for(i=0; i<14; i++){
		for(j=0; j<64; j++){
			for(k=0; k<2; k++){
				pvt[i][j][k] = 0;
			}
		}
	}

	/* Pawns */
	for(i=0; i<64; i++){
		pvt[Pawn][i][Opening] += Pawn_File[file(i)] * Pawn_File_Opening;
		pvt[Pawn][i][Endgame] += Pawn_Rank[rank(i)] * Pawn_Rank_Endgame;
	}

	/* Center control */
	pvt[Pawn][d3][Opening] += 10;
	pvt[Pawn][e3][Opening] += 10;
	pvt[Pawn][d4][Opening] += 20;
	pvt[Pawn][e4][Opening] += 20;
	pvt[Pawn][d5][Opening] += 10;
	pvt[Pawn][e5][Opening] += 10;

	/* Knights */
	for(i=0; i<64; i++){
		pvt[Knight][i][Opening] += Knight_Line[file(i)] * Knight_Center_Opening;
		pvt[Knight][i][Opening] += Knight_Line[rank(i)] * Knight_Center_Opening;
		pvt[Knight][i][Opening] += Knight_Rank[rank(i)] * Knight_Rank_Opening;
		pvt[Knight][i][Endgame] += Knight_Line[file(i)] * Knight_Center_Endgame;
		pvt[Knight][i][Endgame] += Knight_Line[rank(i)] * Knight_Center_Endgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pvt[Knight][i][Opening] -= Knight_Back_Rank_Opening;
	/* Trapped */
	pvt[Knight][a8][Opening] -= Knight_Trapped;
	pvt[Knight][h8][Opening] -= Knight_Trapped;

	/* Bishops */
	for(i=0; i<64; i++){
		pvt[Bishop][i][Opening] += Bishop_Line[file(i)] * Bishop_Center_Opening;
		pvt[Bishop][i][Opening] += Bishop_Line[rank(i)] * Bishop_Center_Opening;
		pvt[Bishop][i][Endgame] += Bishop_Line[file(i)] * Bishop_Center_Endgame;
		pvt[Bishop][i][Endgame] += Bishop_Line[rank(i)] * Bishop_Center_Endgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pvt[Bishop][i][Opening] -= Bishop_Back_Rank_Opening;
	/* Main diagonals*/
	for(i=0; i<8; i++){
		j = getSquare(i,i);
		pvt[Bishop][j][Opening] += Bishop_Diagonal_Opening;
		pvt[Bishop][opposite(j)][Opening] += Bishop_Diagonal_Opening;
	}

	/* Rooks */
	for(i=0; i<64; i++) pvt[Rook][i][Opening] += Rook_File[file(i)] * Rook_File_Opening;

	/* Queens */
	for(i=0; i<64; i++){
		pvt[Queen][i][Opening] += Queen_Line[file(i)] * Queen_Center_Opening;
		pvt[Queen][i][Opening] += Queen_Line[rank(i)] * Queen_Center_Opening;
		pvt[Queen][i][Endgame] += Queen_Line[file(i)] * Queen_Center_Endgame;
		pvt[Queen][i][Endgame] += Queen_Line[rank(i)] * Queen_Center_Endgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pvt[Queen][i][Opening] -= Queen_Back_Rank_Opening;

	/* Kings */
	for(i=0; i< 64; i++){
		pvt[King][i][Opening] += King_File[file(i)] * King_File_Opening;
		pvt[King][i][Opening] += King_Rank[rank(i)] * King_Rank_Opening;
		pvt[King][i][Endgame] += King_Line[file(i)] * King_Center_Endgame;
		pvt[King][i][Endgame] += King_Line[rank(i)] * King_Center_Endgame;
	}

	/* Symmetry copy for black */
	for (i = 0; i < 14; i += 2) {
		for (j = 0; j < 64; j++) {
			for (k= 0; k < 2; k++) {
				pvt[i+1][j][k] = -pvt[i][opposite(j)][k];
         		}
      		}
   	}

	/* Print pvt */
	/*for (k= 0; k < 2; k++) {
		for (i = 0; i < 14; i ++) {
			g_print("%c:\n",pieceToChar[i]);
			for (j = 0; j < 64; j++) {
				if(j%8==0) g_print("\n");
				g_print("%d ",pvt[i][j][k]);
			}
			g_print("\n");
      		}
		g_print("\n");
   	}*/
}

void initAttack(void){
	gint s, b, q, f, r; /* square, bit, diagonal, file, rank*/
	guint64 mask;
	for (s=0;s<64;s++) { /* pre-compute attack masks for w,b pawns: two side squares */
		mask = 1ull<<s;
		attackPawn[White][s]  = shiftUpLeft(mask)   | shiftUpRight(mask);
		attackPawn[Black][s]  = shiftDownLeft(mask) | shiftDownRight(mask);
	}
	for(s=0; s<64; s++){ /** pre-compute 8 knight's attack squares */
		mask = 1ull<<s;
		attackKnight[s] = shiftLeft(shift2Up(mask))    | shiftRight(shift2Up(mask))   |
				  shiftLeft(shift2Down(mask))  | shiftRight(shift2Down(mask)) |
				  shift2Left(shiftUp(mask))    | shift2Right(shiftUp(mask))   |
				  shift2Left(shiftDown(mask))  | shift2Right(shiftDown(mask));
	}
	for (s=0;s<64;s++) {
		for (b=0;b<64;b++) { 
			mask = 0;
			q = s;
			while (file(q)>0 && rank(q)<7) {
				q += 7;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateL45[q])>>shiftL45[s])) break;
			}
			q = s;
			while (file(q)<7 && rank(q)>0) {
				q -= 7;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateL45[q])>>shiftL45[s])) break;
			}
			attackBishopL45[s][b] = mask;
			mask = 0;
			q = s;
			while (file(q)<7 && rank(q)<7) {
				q += 9;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateR45[q])>>shiftR45[s])) break;
			}
			q = s;
			while (file(q)>0 && rank(q)>0) {
				q -= 9;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateR45[q])>>shiftR45[s])) break;
			}
			attackBishopR45[s][b] = mask;
		}
	}
	for (s=0;s<64;s++) {
		for (b=0;b<64;b++) {
			f = file(s);
			q = s+1;
			while (++f<8) {
				attackRookL00[s][b]|=1ull<<q;
				if ((1<<f)&(b<<1)) break;
				++q;
			}
			f = file(s);
			q = s-1;
			while (--f>=0) {
				attackRookL00[s][b]|= 1ull<<q;
				if ((1<<f)&(b<<1)) break;
				--q;
			}
			r = rank(s);
			q = s+8;
			while (++r<8) {
				attackRookL90[s][b] |= 1ull<<q;
				if ((1<<(7-r))&(b<<1)) break;
				q+=8;
			}
			r = rank(s);
			q = s-8;
			while (--r>=0) {
				attackRookL90[s][b] |= 1ull<<q;
				if ((1<<(7-r))&(b<<1)) break;
				q-=8;
			}
		}
	}
	for (s=0;s<64;s++) {
	 	mask = 1ull<<s;
	 	attackKing[s] = shiftLeft(mask) | shiftRight(mask) |
				shiftUp(mask)   | shiftDown(mask)  |
				shiftUpLeft(mask) | shiftUpRight(mask) |
				shiftDownLeft(mask) | shiftDownRight(mask);
	}

	for (s=0;s<64;s++) {
		maskBishop[s] = attackBishopR45[s][0] | attackBishopL45[s][0];
		maskRook[s] = attackRookL00[s][0] | attackRookL90[s][0];
		maskQueen[s] =	maskBishop[s] | maskRook[s];
		maskKnight[s] = attackKnight[s];
	}

	for(s=0; s<64; s++){
		mask = attackKing[s];
		kingSafeDistance1[s] = mask;
		kingSafeDistance2[s] = mask | shiftUpLeft(mask) | shiftUpRight(mask) |
								shiftDownLeft(mask) | shiftDownRight(mask);
	}
}

guint iteratedBitCount(guint n){
    guint count=0;    
    while (n){
        count += n & 0x1u ;    
        n >>= 1 ;
    }
    return count ;
}

void initBitsIn16bits(void){
    guint i ;    
    for (i = 0; i < (0x1u<<16); i++)
        bitsIn16bits[i] = iteratedBitCount(i);
    return ;
}

